; Copyright (c) 2006, Theo Brauers, Forschungszentrum Juelich GmbH ICG-II
; All rights reserved. Unauthorized reproduction prohibited.
; This software may be used, copied, or redistributed as long as it is not
; sold and this copyright notice is reproduced on each copy made.  This
; routine is provided as is without any express or implied warranties
; whatsoever. Email bug reports to th.brauers@fz-juelich.de
;
;+
; USERLEVEL:
;   TOPLEVEL
;
; NAME:
;   write_edf
;
; PURPOSE:
;   This procedure writes an ICG structure into an edf file
;
; CATEGORY:
;   DATAFILES/ZCHN
;
; CALLING SEQUENCE:
;   write_edf, filename, struct
;
; INPUTS:
;   Struct:  data structure
;   Filename: output file name
;
; OPTIONAL INPUTS:
;   Tagnames:   string array of tagnames to be written
;               if not present all tags will written
;
; KEYWORD PARAMETERS:
;   ERROR:  returns 0 if succesfull
;   /STDEV:  if set the stdev of all parameters (if present) will be written
;   /INTERVAL_T:  if set the timeinterval of the time axis will be written
;   ADDITIONAL_HEADER:  if supplied this text will be added to the header of the outfile
;	/QUIET:	if set no info messages will be printed
;	/VERBOSE:	print all messages
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;   Written by: Theo Brauers 2006-Nov-14, as acopy of write_enz
; 	Modified: Theo Brauers 2007-6-26, all edf tags are written for parameter
;									if institute is not given organisation will written instead
; 	Modified: Theo Brauers 2007-7-2, mission is written into MISSION=
;-
PRO write_edf, filename, icg_str, tagnames $
	, STDEV=stdev, INTERVAL_T=interval_t $
    , ADDITIONAL_HEADER=add_header $
    , QUIET=quiet, FILL_VALUE = fill_value $
	, ERROR=err, VERBOSE=verb, _EXTRA= e

IF NOT KEYWORD_SET(quiet) THEN IF KEYWORD_SET(verb) THEN MESSAGE, /INFO, 'new version 2007-6-30'

err = 1
IF N_ELEMENTS(fill_value) EQ 1 THEN fill_v = DOUBLE(fill_value) ELSE fill_v = -9.9900D9

; some error handling
IF NOT is_structure(icg_str) THEN RETURN
IF NOT is_string(filename) THEN RETURN
IF N_ELEMENTS(filename) NE 1 THEN RETURN

header1 = [ '!! FILE '+filename, '!! CREATED ON '+SYSTIME(), $
    '!! BY '+user_at_host() ,'!! WITH '+ file_version() ,'']
IF N_ELEMENTS(add_header) GT 0 THEN header1=[header1, add_header, '']
IF KEYWORD_SET(template) THEN header1=['!! TEMPLATE FOR EDF FILE !!', header1]

IF is_tag(icg_str, '!GLOBAL') THEN BEGIN
    header=[header1, 'X_HEADER=NETCDF_GLOBAL']
    IF is_tag(icg_str.!global, 'PI') THEN BEGIN
        IF is_tag(icg_str.!global.pi, 'NAME') THEN  header= [header, 'PI_NAME='+icg_str.!global.pi.name] $
        	ELSE  header= [header, 'PI_NAME=?? (required!)']
        IF is_tag(icg_str.!global.pi, 'WORKGROUP') THEN  header= [header, 'WORKGROUP='+icg_str.!global.pi.workgroup]
        IF is_tag(icg_str.!global.pi, 'INSTITUTE') THEN  header= [header, 'INSTITUTE='+icg_str.!global.pi.institute] $
        ELSE BEGIN
	        IF is_tag(icg_str.!global.pi, 'ORGANISATION') THEN  header= [header, 'INSTITUTE='+icg_str.!global.pi.organisation]
        	header= [header, 'INSTITUTE=?? (required!)']
        ENDELSE
        IF is_tag(icg_str.!global.pi, 'ORGANISATION') THEN  header= [header, 'ORGANISATION='+icg_str.!global.pi.organisation]
        IF is_tag(icg_str.!global.pi, 'EMAIL') THEN  header= [header, 'EMAIL='+icg_str.!global.pi.email] $
        	ELSE  header= [header, 'EMAIL=?? (required!)']
        IF is_tag(icg_str.!global.pi, 'ADDRESS') THEN  header= [header, 'ADDRESS='+icg_str.!global.pi.address]
        IF is_tag(icg_str.!global.pi, 'PHONE') THEN  header= [header, 'PHONE='+icg_str.!global.pi.phone]
    ENDIF ELSE header= [header, 'PI_NAME=?? (required!)','INSTITUTE=?? (required!)','EMAIL=?? (required!)']
    IF is_tag(icg_str.!global, 'DATASET') THEN BEGIN
        IF is_tag(icg_str.!global.dataset, 'TITLE') THEN  header= [header, 'TITLE='+icg_str.!global.dataset.title] $
        	ELSE  header= [header, 'TITLE=?? (required!)']
        IF is_tag(icg_str.!global.dataset, 'DATA_CATEGORY') THEN  header= [header, 'DATA_CATEGORY='+icg_str.!global.dataset.data_category] $
        	ELSE  header= [header, 'DATA_CATEGORY=?? (required!)']
        IF is_tag(icg_str.!global.dataset, 'EXPERIMENT') THEN  header= [header, 'EXPERIMENT='+icg_str.!global.dataset.experiment]
        IF is_tag(icg_str.!global.dataset, 'CAMPAIGN') THEN  header= [header, 'CAMPAIGN='+icg_str.!global.dataset.campaign]
        IF is_tag(icg_str.!global.dataset, 'MISSION') THEN  header= [header, 'MISSION='+icg_str.!global.dataset.mission]
        IF is_tag(icg_str.!global.dataset, 'TYPE_OF_DATA') THEN  header= [header, 'TYPE_OF_DATA='+icg_str.!global.dataset.type_of_data] $
        	ELSE  header= [header, 'TYPE_OF_DATA=?? (required!)']
        IF is_tag(icg_str.!global.dataset, 'STATUS_OF_FILE') THEN  header= [header, 'STATUS_OF_FILE='+icg_str.!global.dataset.status_of_file] $
        	ELSE  header= [header, 'STATUS_OF_FILE=?? (required!)']
        IF is_tag(icg_str.!global.dataset, 'VERSION') THEN  header= [header, 'VERSION='+icg_str.!global.dataset.version]
        IF is_tag(icg_str.!global.dataset, 'SOURCE_FILE') THEN  header= [header, 'SOURCE_FILE='+icg_str.!global.dataset.source_file]
        IF is_tag(icg_str.!global.dataset, 'DATA_ORIGIN') THEN  header= [header, 'DATA_ORIGIN='+icg_str.!global.dataset.data_origin]
    ENDIF ELSE header= [header, 'PI_NAME=?? (required!)','INSTITUTE=?? (required!)','EMAIL=?? (required!)']
    IF is_tag(icg_str.!global, 'PLATFORM') THEN BEGIN
        IF is_tag(icg_str.!global.platform, 'TYPE') THEN  header= [header, 'PLATFORM='+icg_str.!global.platform.type]$
        	ELSE  header= [header, 'PLATFORM=?? (required!)']
        IF is_tag(icg_str.!global.platform, 'NAME') THEN  header= [header, 'NAME_OF_PLATFORM='+icg_str.!global.platform.name] $
        	ELSE  header= [header, 'NAME_OF_PLATFORM=?? (required!)']
    ENDIF ELSE header= [header, 'PLATFORM=?? (required!)','NAME_OF_PLATFORM=?? (required!)']
ENDIF

header=[header, '', 'X_HEADER=NETCDF_TIME']

names=get_tag_and_short_names(icg_str, count)
IF N_ELEMENTS(tagnames) GT 0 THEN tag_n2 = tagnames ELSE tag_n2 = names.tag_name
IF N_ELEMENTS(stdev) NE N_ELEMENTS(tag_n2) THEN $
	IF KEYWORD_SET(stdev) THEN stdev2 = MAKE_ARRAY(N_ELEMENTS(tag_n2), /BYTE, VAL=1B) $
	ELSE stdev2 = MAKE_ARRAY(N_ELEMENTS(tag_n2), /BYTE, VAL=0B) $
ELSE stdev2 = stdev

text=''
iy = tag_position(icg_str, tag_n2, count)
IF count LT 1 THEN RETURN
it = tag_position(icg_str, 'TIME', ct)
IF ct GT 1 THEN RETURN
k=0
IF ct EQ 1 THEN BEGIN
    tvalue = icg_str.(it).param
    text = [text, 'TIME' ]
    IF is_tag(icg_str.(it), 'SHORT_NAME') THEN  header= [header, 'SHORT_NAME'+'('+n2s(k)+')=TIME' ]
    IF is_tag(icg_str.(it), 'LONG_NAME') THEN  header= [header, 'LONG_NAME'+'('+n2s(k)+')=TIME' ]
    IF is_tag(icg_str.(it), 'UNITS') THEN  header= [header, 'UNITS'+'('+n2s(k)+')=seconds since 2000-1-1 0:00:00 UTC' ]
    IF is_tag(icg_str.time, 'INTERVAL_T') AND KEYWORD_SET(interval_t) THEN BEGIN
        tvalue = [[tvalue], [icg_str.(it).interval_t]]
        text = [text, 'INTERVAL']
    ENDIF
ENDIF

iy = a_not_b(a_not_b(iy, it), [-1L])
count = N_ELEMENTS(iy)
header=[header, '', 'X_HEADER=NETCDF_PARAMETER']
FOR i=0, count-1 DO BEGIN
    CASE 1 OF
        ((SIZE(icg_str.(iy[i]).param))[0] NE 1 AND N_ELEMENTS(icg_str.(iy[i]).param) NE 1): $
            MESSAGE, 'Cannot write 2 and more dim parameters: ' + icg_str.(iy[i]).short_name, /INFO
        (TOTAL(SIZE(icg_str.(iy[i]).param, /TYPE) EQ [0,6,7,8,9,10,11]) EQ 1): $
            MESSAGE, 'Cannot write non num parameters: ' + icg_str.(iy[i]).short_name, /INFO
        ELSE: BEGIN
        	temp = DOUBLE(icg_str.(iy[i]).param)
        	idxn = WHERE(FINITE(temp, /NAN), cdxn)
        	IF cdxn GT 0 THEN BEGIN
            	IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
            	temp[idxn] = fill_v
        	ENDIF
            text = [text, text2tagname( icg_str.(iy[i]).short_name)]
           	dvalue = (N_ELEMENTS(dvalue) EQ 0) ? temp : [[dvalue], [temp]]
            IF KEYWORD_SET(is_tag(icg_str.(iy[i]), 'STDEV')) AND KEYWORD_SET(stdev2[i]) THEN BEGIN
            	temp = DOUBLE(icg_str.(iy[i]).stdev)
				idxn = WHERE(FINITE(temp, /NAN), cdxn)
        		IF cdxn GT 0 THEN BEGIN
            		IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN icg_str.(iy[i]).missing_value = fill_v
            		temp[idxn] = fill_v
        		ENDIF
           	    dvalue = [[dvalue], [temp]]
                text = [text, 'STDEV('+text2tagname( icg_str.(iy[i]).short_name)+')']
            ENDIF
            header= [header, '','LONG_NAME'+'('+n2s(incr(k))+')='+icg_str.(iy[i]).long_name ]
			header= [header, 'SHORT_NAME'+'('+n2s(k)+')='+text2tagname( icg_str.(iy[i]).short_name )]
            IF is_tag(icg_str.(iy[i]), 'MISSING_VALUE') THEN $
            	header= [header, 'MISSING_VALUE'+'('+n2s(k)+')='+n2s(icg_str.(iy[i]).missing_value) ] $
            ELSE BEGIN
             	IF is_tag(icg_str.(iy[i]), 'FILL_VALUE') THEN $
	            	header= [header, 'MISSING_VALUE'+'('+n2s(k)+')='+n2s(icg_str.(iy[i]).fill_value) ] $
				ELSE $
            		header= [header, 'MISSING_VALUE'+'('+n2s(k)+')='+n2s(fill_v) ]
            ENDELSE
            IF is_tag(icg_str.(iy[i]), 'UNITS') THEN $
            	header= [header, 'UNITS'+'('+n2s(k)+')='+icg_str.(iy[i]).units ]
            IF is_tag(icg_str.(iy[i]), 'DESCRIPTION') THEN $
            	header= [header, 'DESCRIPTION'+'('+n2s(k)+')='+icg_str.(iy[i]).description ]
            IF is_tag(icg_str.(iy[i]), 'ACCURACY_ABS') THEN $
            	header= [header, 'ACCURACY_ABS'+'('+n2s(k)+')='+n2s(icg_str.(iy[i]).accuracy_abs) ]
            IF is_tag(icg_str.(iy[i]), 'ACCURACY_REL') THEN $
            	header= [header, 'ACCURACY_REL'+'('+n2s(k)+')='+n2s(icg_str.(iy[i]).accuracy_rel) ]
            IF is_tag(icg_str.(iy[i]), 'DETECTION_LIMIT') THEN $
            	header= [header, 'DETECTION_LIMIT'+'('+n2s(k)+')='+n2s(icg_str.(iy[i]).detection_limit) ]
            IF is_tag(icg_str.(iy[i]), 'INSTRUMENT') THEN $
            	header= [header, 'INSTRUMENT'+'('+n2s(k)+')='+icg_str.(iy[i]).instrument ]
            IF is_tag(icg_str.(iy[i]), 'CALIBRATION') THEN $
            	header= [header, 'CALIBRATION'+'('+n2s(k)+')='+icg_str.(iy[i]).calibration ]
            IF is_tag(icg_str.(iy[i]), 'INSTRUMENT_FILE') THEN $
            	header= [header, 'INSTRUMENT_FILE'+'('+n2s(k)+')='+icg_str.(iy[i]).instrument_file ]
            IF is_tag(icg_str.(iy[i]), 'ANALYSIS_FILE') THEN $
            	header= [header, 'ANALYSIS_FILE'+'('+n2s(k)+')='+icg_str.(iy[i]).analysis_file ]
        END
    ENDCASE
ENDFOR

IF N_ELEMENTS(dvalue) EQ 0 THEN dvalue = tvalue
IF N_ELEMENTS(dvalue) GT 0 AND N_ELEMENTS(tvalue) GT 0 THEN dvalue = [[tvalue],[dvalue]]

sd = SIZE(dvalue)
IF N_ELEMENTS(sd) EQ 5 THEN BEGIN & ix = sd[1] & iy = sd[2] & ENDIF
IF N_ELEMENTS(sd) EQ 4 THEN BEGIN & ix = 1 & iy = sd[1] & ENDIF
IF N_ELEMENTS(sd) EQ 3 THEN BEGIN & ix = 1 & iy = 1 & ENDIF

h2 = STRTRIM(str_sep2(header, '=', ONLY=1),2)
idx = WHERE(STRPOS(h2, '??') EQ 0,cdx)
IF cdx NE 0 AND NOT KEYWORD_SET(quiet) THEN FOR i=0,cdx-1 DO MESSAGE, /INFO, 'missing: '+header[idx[i]]

header = [header, '', 'X_HEADER=ENZ', make_enz_header(text[1:*])]
iwzn, filename, dvalue, ix, iy, header, err, _EXTRA = e

RETURN
END
